/*

 normal.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "normal.h"
#include "driver.h"
#include "ideservd.h"
#include "path.h"
#include "crc8.h"
#include "partition.h"
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

static inline int send_trailer(int usecrc, const struct io_ops_s *io_ops, const struct arguments_s *arguments) {
    if (arguments->mode != M_ETHERNET) {
        if (usecrc) {
            if (io_ops->sendb(crc_get())) return 1;
        }
        if (io_ops->sendb(0x5a)) return 1;
    }
    return io_ops->flush();
}

static int check_trailer(const char *mode, int usecrc, const struct io_ops_s *io_ops, FILE *lf) {
    int fr;
    if (usecrc) {
        if (io_ops->getb() < 0) return 1;
        if (crc_get()) {
	    seterror(42, 1);
            fprintf(lf, "%s: CRC error\n", mode);
            io_ops->getb();
            return 1;
        }
    }
    fr = io_ops->getb();if (fr < 0) return 1;
    if (fr != 0x5a) {
        seterror(41,1);
	fprintf(lf, "%s: Frame error %02x\n", mode, fr);
	return 1;
    }
    return 0;
}

int statuserror(const struct io_ops_s *io_ops, const char *errorbuff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    c64c_t cmd[256];
    unsigned char fp;
    int r;

    if (io_ops->getb() < 0) return 1;
    fp = 0;
    while (fp<255) {
        r = io_ops->getb();
        if (r < 0) {
            if (arguments->mode == M_ETHERNET) break;
            return 1;
        }
        cmd[fp] = r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp] = 0;

    if (arguments->mode != M_ETHERNET && check_trailer("Status", usecrc, io_ops, lf)) return 1;

    alarm(0);

    if (arguments->debug) {
        char msg[16 * 3];
        int i=0;
        fprintf(lf, "Command:\n");
        while (cmd[i]) {
            snprintf(msg + (i & 15) * 3, 3, "%02x ", cmd[i]);
            i++;
            if (!(i & 15)) {
                msg[16 * 3 - 1] = '\0';
                fprintf(lf, "%s\n", msg);
            }
        }
        if (i & 15) {
            msg[(i & 15) * 3 - 1] = '\0';
            fprintf(lf, "%s\n", msg);
        }
    }
    commandchannel(cmd);

    crc_clear();
    fp = strlen(errorbuff);
    if (io_ops->sendb(fp)) return 1;
    if (arguments->mode == M_ETHERNET && io_ops->sendb(fp)) return 1;
    if (io_ops->sendbytes(fp, (unsigned char *)errorbuff)) return 1;
    if (send_trailer(usecrc, io_ops, arguments)) return 1;
    return 0;
}

int openfile(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    c64c_t cmd[256];
    unsigned char status=0x80;
    unsigned char mode='R';
    unsigned char channel,fp;
    unsigned long long l=0;
    unsigned int length=0;
    struct buffer_s *buffer;
    int r;

    r = io_ops->getb();
    if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];

    fp=0;
    while (fp < 255) {
        r = io_ops->getb();
        if (r < 0) {
            if (arguments->mode == M_ETHERNET) break;
            return 1;
        }
        cmd[fp] = r;
        if (!r) break;
        fp++;
    }
    cmd[fp]=0;

    if (arguments->mode != M_ETHERNET && check_trailer("Open", usecrc, io_ops, lf)) return 1;

    alarm(0);

    if (arguments->debug) {
        char msg[16 * 3];
        int i=0;
        fprintf(lf, "Command:\n");
        while (cmd[i]) {
            snprintf(msg + (i & 15) * 3, 3, "%02x ", cmd[i]);
            i++;
            if (!(i & 15)) {
                msg[16 * 3 - 1] = '\0';
                fprintf(lf, "%s\n", msg);
            }
        }
        if (i & 15) {
            msg[(i & 15) * 3 - 1] = '\0';
            fprintf(lf, "%s\n", msg);
        }
    }
    if (buffer->file) {
        fclose(buffer->file);
        buffer->file=NULL;
    }
    buffer->mode = CM_CLOSED;

    if (channel == 15) {
        buffer->mode = CM_ERR;
    } else if (cmd[0] == '$') {
        unsigned char dirline[32];
        struct dirent *ep;
        DIR *dp;
        int sum=0,partition;
        char outpath[1000];
        buffer->mode=CM_DIR;
        buffer->length=buffer->pointer=0;
        c64c_t name[17], type[4]={'*',0};
        const c64c_t *outname;
        partition_t outpart = 0;

        partition=!strncmp((char *)&cmd[1],"=P",2);

        resolv_path(cmd+1, outpath, &outname, &outpart, lf, arguments->relaxfiletypes);
        if (arguments->verbose) fprintf(lf, "Open [%d]: \"$%s\"\n", channel, outpath);
        convertfilename(outname, '=', name, type, NULL);

        if (!name[0]) strcpy((char *)name, "*");

        if (channel==1) {
            status = 0x80 | 70;goto vege; //no channel
        } else if (channel>1) {
            dp=opendir(outpath[0]?outpath:".");
            if (!dp) {
                fprintf(lf, "Open: Couldn't open the directory \"%s\": %s(%d)\n", outpath, strerror(errno), errno);
                status = 0x80 | 62;goto vege;
            }
            if (!buffer->data) {
                buffer->data=(unsigned char *)malloc(65536);
                if (!buffer->data) exit(1);//out of memory?!
            }
            buffer->data[buffer->length++]='I';

            memset(&dirline[0],32,16);
            memset(&dirline[16],0,8);
            dirline[24]=A_DIR;
            memcpy(&dirline[25],"DIR",3);
            memset(&dirline[28],0,4);
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

            initdirentry(0);
            while ((ep=readdir(dp)))
            {
                char lname2[2020];

                strcpy(lname2,outpath);
                if (outpath[0]) strncat(lname2,"/",999);
                strncat(lname2,ep->d_name,999);

                if (createdirentry(ep->d_name, lname2, (directory_entry_t*)dirline, arguments->relaxfiletypes)) continue;

                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            }
            initdirentry(1);
            memset(buffer->data + buffer->length, 0, (-buffer->length) & 511);

            status = 0x80 | errtochannel15(closedir(dp));
        } else if (partition) {
            int i, n=0;
            memcpy(&dirline[0x00],"\1\4\1\1\377\0\22\42",8);
            memcpy(&dirline[0x08],"IDESERVD " VERSION2 "  ",16);
            memcpy(&dirline[0x18],"\42\40" SYSTEMNAME "\0",8);
            if (!buffer->data) {
                buffer->data=(unsigned char *)malloc(65536);
                if (!buffer->data) exit(1);//out of memory?!
            }
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            for (i=1;i<255;i++) {
                c64c_t *name = partition_get_name(i);
                if (!name) continue;
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=i;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"   \42\42                 NET  \0",28);
                memcpy(&dirline[0x08], name, strlen((char *)name));
                dirline[0x08 + strlen((char *)name)] = '"';
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                n++;
            }
            dirline[0x00]=dirline[0x01]=1;
            dirline[0x02]=n;
            dirline[0x03]=0;
            memcpy(&dirline[0x04],"PARTITIONS.              \0\0\0",28);
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            memset(buffer->data + buffer->length, 0, (-buffer->length) & 511);
            status = 0x80 | errtochannel15(0);
        } else {

            dp=opendir(outpath[0]?outpath:".");

            if (!dp) {
                fprintf(lf, "Open: Couldn't open the directory \"%s\": %s(%d)\n", outpath, strerror(errno), errno);
                status = 0x80 | 62;goto vege;
            }

            memcpy(&dirline[0x00],"\1\4\1\1\1\0\22\42",8);
            dirline[0x04] = outpart ? outpart : partition_get_current();
            memset(&dirline[0x08],32,16);
            memcpy(&dirline[0x18],"\42\40" SYSTEMNAME "\0",8);
            if (!buffer->data) {
                buffer->data=(unsigned char *)malloc(65536);
                if (!buffer->data) exit(1);//out of memory?!
            }
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

            initdirentry(0);
            while ((ep=readdir(dp)))
            {
                directory_entry_t dirent;
                int offs;
                unsigned int n;
                char lname2[2020];

                strcpy(lname2,outpath);
                if (outpath[0]) strncat(lname2,"/",999);
                strncat(lname2,ep->d_name,999);

                if (createdirentry(ep->d_name, lname2, &dirent, arguments->relaxfiletypes)) continue;

                if (dirent.pointer[0] & 0x80) continue;
                n=(dirent.filesize[0]+(dirent.filesize[1] << 8)+(dirent.filesize[2] << 16)+(dirent.filesize[3] << 24)+255) >> 8;
                if (n>65535) n=65535;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name, name)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=n;
                dirline[0x03]=n >> 8;
                sum+=n;
                memset(&dirline[0x04],32,27);
                offs=7;
                if (n>9) offs--;
                if (n>99) offs--;
                if (n>999) offs--;
                dirline[offs++]=0x22;
                n=0;while (dirent.name[n] && n<16) dirline[offs++]=dirent.name[n++];
                dirline[offs++]=0x22;
                offs+=(16-n);

                dirline[offs++]=(dirent.attrib & A_CLOSED)?' ':'*';

                n=0;while (n<3 && dirent.filetype[n]) {dirline[offs+n]=dirent.filetype[n];n++;}
                if ((dirent.attrib & (A_DELETEABLE | A_WRITEABLE))!=(A_DELETEABLE | A_WRITEABLE)) dirline[offs+3]='<';

                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                if (sum>65535) sum=65535;
            }
            initdirentry(1);

            dirline[0x00]=dirline[0x01]=1;
            dirline[0x02]=sum;
            dirline[0x03]=sum >> 8;
            memcpy(&dirline[0x04],"BLOCKS USED.             \0\0\0",28);

            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            memset(buffer->data + buffer->length, 0, (-buffer->length) & 511);
            status = 0x80 | errtochannel15(closedir(dp));
        }
        length=buffer->length;
    } else {
        c64c_t name[17], type[4]={'*',0};
        const c64c_t *outname;
        unsigned char overwrite;
        char lname[1000],outpath[1020];
        struct dirent *ep;
        DIR *dp;
        int found=0;
        int b;

        overwrite=0;
        for (b=0;cmd[b];b++) if (cmd[b]==':') {
            if (cmd[0]=='@') overwrite=1;
            break;
        }

        resolv_path(cmd, outpath, &outname, NULL, lf, arguments->relaxfiletypes);
        if (channel==1) strcpy((char *)type,"PRG");
        convertfilename(outname, ',', name, type, &mode);

        dp=opendir(outpath[0]?outpath:".");
        if (!dp) {
            fprintf(lf, "Open: Couldn't open the directory \"%s\": %s(%d)\n", outpath, strerror(errno), errno);
            status = 0x80 | 62;goto vege;
        }

        initdirentry(0);
        while ((ep=readdir(dp)))
        {
            directory_entry_t dirent;
            char lname2[2020];

            strncpy(lname,ep->d_name,999);
            strcpy(lname2,outpath);
            if (outpath[0]) strncat(lname2,"/", 999);
            strncat(lname2,ep->d_name,999);

            if (createdirentry(ep->d_name, lname2, &dirent, arguments->relaxfiletypes)) continue;

            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

            length = dirent.filesize[0] + dirent.filesize[1] * 256 + dirent.filesize[2] * 65536 + dirent.filesize[3] * 65536 * 256;

            dirent.filesize[0]=0;
            if (!matchname(dirent.name, name)) continue;
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;

            found=1;
            break;
        }
        closedir(dp);
        initdirentry(1);

        if (!found) {
            mbstate_t ps;
            int f=0,f2=0,g=0;
            memset(&ps, 0, sizeof(ps));
            while (name[f] && f < 16) {
                f2 += c64toascii(lname + f2, name[f], &ps);
                f++;
            }
            if (!arguments->relaxfiletypes || memcmp(type,"PRG",3)) {
                memset(&ps, 0, sizeof(ps));
                lname[f2++]=',';
                while (type[g] && g < 3) {
                    f2 += c64toascii(lname + f2, type[g], &ps);
                    g++;
                }
            }
            lname[f2]=0;
        }

        if (channel==0) mode='R';
        if (channel==1) mode='W';

        if (outpath[0]) strcat(outpath,"/");
        strcat(outpath,lname);

        if (arguments->verbose) fprintf(lf,"Open [%d]: \"%s\" %c\n", channel, outpath, mode);

        if (!name[0]) {
            status=0x80 | 34;goto vege;//no filename
        }

        if (mode=='W') {
            int f=0;
            length=0;

            while (lname[f]) {
                if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                    status=0x80 | 33;goto vege;//illegal characters
                }
                f++;
            }

            if (found && !overwrite) {
                status=0x80 | 63;//exists
            } else {
#ifndef WIN32
                buffer->fd=open(outpath, O_CREAT|O_RDWR|(overwrite ? O_TRUNC : O_EXCL), S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
#else
                buffer->fd=open(outpath, O_CREAT|O_RDWR|(overwrite ? O_TRUNC : O_EXCL)|O_BINARY, S_IRUSR|S_IWUSR);
#endif
                if (buffer->fd<0) status = 0x80 | errtochannel15(1); else {
                    status=0x80;
                    buffer->file=fdopen(buffer->fd,"wb+");
                    buffer->mode=CM_FILE;
                    buffer->filesize=0;
                }
            }
        } else {
            if (!found) {
                status=0x80 | 62;//not found
            } else {
                if (mode=='R') {
                    buffer->fd=open(outpath, O_RDONLY|O_BINARY, 0);
                    if (buffer->fd<0) status = 0x80 | errtochannel15(1); else {
                        status=0x80;//ok
                        buffer->file=fdopen(buffer->fd, "rb");
                        buffer->mode=CM_FILE;
                        buffer->filesize=length;
                    }
                }
                if (mode=='A' || mode=='M') {
                    buffer->fd=open(outpath, O_RDWR|O_BINARY, 0);
                    if (buffer->fd<0) status = 0x80 | errtochannel15(1); else {
                        status=0x80;//ok
                        buffer->file=fdopen(buffer->fd, "rb+");
                        buffer->mode=CM_FILE;
                        buffer->filesize=length;
                    }
                }
            }
        }
    }
vege:
    {
        unsigned char buf[6];
        crc_clear();
        if (arguments->mode == M_ETHERNET) {
            buf[0] = length;
            buf[1] = length >> 8;
            buf[2] = length >> 16;
            buf[3] = length >> 24;
            buf[4] = mode;
            buf[5] = status;
        } else {
            buf[0] = status;
            buf[1] = mode;
            buf[2] = length;
            buf[3] = length >> 8;
            buf[4] = length >> 16;
            buf[5] = length >> 24;
        }
        if (io_ops->sendbytes(sizeof(buf), buf)) return 1;
        if (send_trailer(usecrc, io_ops, arguments)) return 1;
    }
    return 0;
}

int closefile(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    unsigned char channel;
    unsigned int length;
    struct buffer_s *buffer;
    unsigned char buf[5];

    if (io_ops->getbytes(sizeof(buf), buf)) return 1;
    channel=buf[0] & 0x0f;
    length=(((((buf[4] << 8) | buf[3]) << 8) | buf[2]) << 8) | buf[1];
    if (arguments->mode != M_ETHERNET && check_trailer("Close", usecrc, io_ops, lf)) return 1;
    buffer=&buff[channel];

    alarm(0);

    if (arguments->verbose) fprintf(lf,"Close [%d] %d byte(s)\n", channel, length);
    if (channel == 15) buffer->mode = CM_ERR;
    else {
        if (buffer->file) {
            if (buffer->mode==CM_FILE && length >= buffer->filesize) {
                fflush(buffer->file);
                if (ftruncate(buffer->fd,length)) {
                    fprintf(lf, "Close: Couldn't truncate: %s(%d)\n", strerror(errno), errno);
                }
            }
            fclose(buffer->file);
        }
        buffer->mode=CM_CLOSED;buffer->file=NULL;
        if (buffer->data) {
            free(buffer->data);
            buffer->data=NULL;
        }
    }
    crc_clear();
    if (io_ops->sendb(0x80)) return 1;
    if (arguments->mode == M_ETHERNET && io_ops->sendb(0x80)) return 1;
    if (send_trailer(usecrc, io_ops, arguments)) return 1;
    return 0;
}

int readfile(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    unsigned char channel, sectors;
    unsigned short int j;
    struct timeval start,end;
    unsigned long long l=0,l2;
    unsigned int address;
    struct buffer_s *buffer;
    int fr, bytes = 0;
    unsigned char buf[5];

    if (io_ops->getbytes(sizeof(buf), buf)) return 1;
    channel=buf[0] & 0x0f;
    address=(((buf[3] << 8) | buf[2]) << 8) | buf[1];
    sectors=buf[4];
    if (arguments->mode != M_ETHERNET && check_trailer("Read", usecrc, io_ops, lf)) return 1;
    if (arguments->mode == M_ETHERNET) {
        if (sectors>2) sectors=2;
    } else {
        if (sectors>128) sectors=128;
    }
    buffer=&buff[channel];

    alarm(0);

    if (arguments->verbose) fprintf(lf,"Read [%d] %06x %d sector(s)\n", channel, address, sectors);
    if (buffer->mode!=CM_DIR && buffer->mode!=CM_ERR && (buffer->mode!=CM_FILE || !buffer->file)) {
        fprintf(lf, (buffer->mode == CM_CLOSED) ? "Read: Channel not open\n" : "Read: Not readable\n");
        crc_clear();
        if (io_ops->sendb(0x80 | 70)) return 1;
        if (arguments->mode == M_ETHERNET && io_ops->sendb(0x80 | 70)) return 1;
        if (send_trailer(usecrc, io_ops, arguments)) return 1;
        return 0;
    }

    if (buffer->mode==CM_FILE) {
	unsigned long itt;
	fseek(buffer->file, address << 8, SEEK_SET);
        if (!buffer->data) {
            buffer->data=(unsigned char *)malloc(65536);
            if (!buffer->data) exit(1);//out of memory?!
        }
        itt=fread(buffer->data,1,sectors*512,buffer->file);

	if (itt<sectors*512) memset(buffer->data+itt,0,sectors*512-itt);
	j=0;
    } else {
        j=buffer->pointer;
    }

    if (arguments->mode == M_ETHERNET) {
        crc_clear();
        if (io_ops->sendbytes(sectors * 512, buffer->data + j)) return 1;
        if (io_ops->sendb(0x80)) return 1;
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc, io_ops, arguments)) return 1;
        buffer->pointer+=sectors*512;
    } else {
        crc_clear();
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc, io_ops, arguments)) return 1;
        if (arguments->verbose) {
            bytes = sectors * 512;
            gettimeofday(&start,NULL);
        }
        while (sectors) {
            if (arguments->mode == M_RS232 || arguments->mode == M_RS232S) {
                fr = io_ops->getb();if (fr < 0) return 1;
                if (fr != 0x5a) {
                    fprintf(lf, "Read: Frame error %02x\n", fr);
                    return 1;
                }
            }
            crc_clear();
            if (io_ops->sendbytes(512,buffer->data+j)) return 1;
            j+=512;
            if (io_ops->sendb(0x80)) return 1;
            if (send_trailer(usecrc, io_ops, arguments)) return 1;
            sectors--;buffer->pointer=j;l+=512;
        }
        if (arguments->verbose) {
            gettimeofday(&end,NULL);
            l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
            if (l2) {
                l=l*1000000/l2;
                fprintf(lf, "Read: Sent %d bytes at %lld bytes/seconds\n", bytes, l);
            } else {
                fprintf(lf, "Read: Sent %d bytes\n", bytes);
            }
        }
    }
    return 0;
}

int writefile(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    unsigned char channel, sectors;
    struct timeval start,end;
    unsigned long long l=0,l2;
    unsigned int address;
    struct buffer_s *buffer;
    int fr, bytes = 0;
    unsigned char buf[5];

    if (io_ops->getbytes(sizeof(buf), buf)) return 1;
    channel=buf[0] & 0x0f;
    address=(((buf[3] << 8) | buf[2]) << 8) | buf[1];
    sectors=buf[4];
    if (arguments->mode != M_ETHERNET && check_trailer("Write", usecrc, io_ops, lf)) return 1;
    if (sectors>128) sectors=128;
    buffer=&buff[channel];

    alarm(0);

    if (arguments->verbose) fprintf(lf, "Write [%d] %06x %d sector(s)\n", channel, address, sectors);
    if (buffer->mode!=CM_FILE || !buffer->file) {
        fprintf(lf, "Write: Not writeable\n");
        crc_clear();
        if (io_ops->sendb(0x80+25)) return 1;
        if (arguments->mode == M_ETHERNET && io_ops->sendb(0x80+25)) return 1;
        if (send_trailer(usecrc, io_ops, arguments)) return 1;
        return 0;
    }
    if (!buffer->data) {
        buffer->data=(unsigned char *)malloc(65536);
        if (!buffer->data) exit(1);//out of memory?!
    }
    fseek(buffer->file, address << 8, SEEK_SET);
    if (arguments->mode == M_ETHERNET) {
        if (sectors>2) sectors=2;
        if (io_ops->getbytes(512*sectors, buffer->data)) return 1;
        crc_clear();
        if (fwrite(buffer->data,1,512*sectors,buffer->file)==512*sectors) {
            if (io_ops->sendb(0x80)) return 1;
            if (io_ops->sendb(0x80)) return 1;
        } else {
            if (io_ops->sendb(0x80+25)) return 1;
            if (io_ops->sendb(0x80+25)) return 1;
        }
        if (send_trailer(usecrc, io_ops, arguments)) return 1;
    } else {
        crc_clear();
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc, io_ops, arguments)) return 1;

        if (arguments->verbose) {
            bytes = sectors * 512;
            gettimeofday(&start,NULL);
        }
        while (sectors) {
    	    crc_clear();
  	    if (io_ops->getbytes(512, buffer->data)) return 1;
            if (usecrc) {
    	        if (io_ops->getb() < 0) return 1;
    	        if (crc_get()) {
    	    	    seterror(42,1);
    	    	    fprintf(lf, "Write: CRC error\n");
    	    	    io_ops->getb();
    	    	    return 1;
    	    	}
	    }
            fr = io_ops->getb();if (fr < 0) return 1;
            if (fr != 0x5a) {
                seterror(41,1);
                fprintf(lf, "Write: Frame error %02x\n", fr);
                return 1;
            }
    	    crc_clear();
    	    if (fwrite(buffer->data,1,512,buffer->file)==512) {
                if (sectors==1 || arguments->mode == M_RS232 || arguments->mode == M_RS232S) {
		    if (io_ops->sendb(0x80)) return 1;
            	    if (send_trailer(usecrc, io_ops, arguments)) return 1;
            	}
    	    } else {
	        if (io_ops->sendb(0x80+25)) return 1;
    	        if (send_trailer(usecrc, io_ops, arguments)) return 1;
    	        break;
    	    }
            sectors--;l+=512;
        }
        if (arguments->verbose) {
            gettimeofday(&end,NULL);
            l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
            if (l2) {
                l=l*1000000/l2;
                fprintf(lf, "Write: Received %d bytes at %lld bytes/seconds\n", bytes, l);
            } else {
                fprintf(lf, "Write: Received %d bytes\n", bytes);
            }
        }
    }
    return 0;
}

